import os
import yaml


class NodeMissingError(KeyError):
    """Raised if data node doesn't exist."""


class QueryTypeError(TypeError):
    """Raised if query is of an invalid type."""


class SourceInvalidError(FileNotFoundError):
    """Raised if the data source cannot be found."""


class ValueCountError(Exception):
    """Raised if query is incomplete."""


class Reader:
    def __init__(self, source: str) -> None:
        yaml_data = os.path.join(os.path.dirname(__file__), "sources.yml")
        try:
            if not os.path.exists(yaml_data):
                raise SourceInvalidError(
                    "Cannot find source at '{}'!".format(yaml_data)
                )
            data_source = yaml.load(open(yaml_data), Loader=yaml.FullLoader)
            if source not in data_source:
                raise NodeMissingError("Cannot find source table '{}'!".format(source))
            self.data_source = data_source[source]
            sorted(self.data_source)
        except SourceInvalidError as error:
            exit(error)
        except NodeMissingError as error:
            exit(error)

    def find(self, query=None) -> (dict, list):
        """Returns results for query, if any."""
        data_source = self.data_source
        if query is None or isinstance(query, (list, tuple)) and len(query) is 0:
            return data_source
        try:
            if not isinstance(query, (list, tuple)):
                raise QueryTypeError(
                    "find: query argument must be type 'list' or 'tuple'!"
                )
            if len(query) == 3:
                data_source = self.data_source[query[0]][query[1]][query[2]]
            elif len(query) == 2:
                data_source = self.data_source[query[0]][query[1]]
            elif len(query) == 1:
                data_source = self.data_source[query[0]]
            else:
                raise ValueCountError(
                    "find: query requires a max of '3' purged_values. {} given.".format(
                        len(query)
                    )
                )
        except ValueCountError as error:
            exit("find: {}".format(error))
        except QueryTypeError as error:
            exit(error)
        except KeyError:
            return dict()
        else:
            if isinstance(data_source, dict):
                sorted(data_source)
            if isinstance(data_source, list):
                data_source.sort()
        return data_source



from collections import OrderedDict
import os
import time

from bs4 import BeautifulSoup
import click

from personaednd.attributes import ScoreAssigner, ScoreImprovement
from personaednd.sources import (
    get_ability_modifier,
    get_proficiency,
    get_proficiency_bonus,
    Backgrounds,
    Classes,
    Races,
    Skills,
)
from personaednd.sources.reader import Reader
from personaednd.version import __version__


class ValidParameters(Reader):
    def __init__(self, source=None) -> None:
        super().__init__(source)

    @classmethod
    def __toTuple__(cls, source, index=None) -> (tuple, str):
        _tuple = tuple(cls(source).find().keys())
        if isinstance(index, int):
            try:
                return _tuple[index]
            except IndexError:
                return _tuple
        return _tuple

    @classmethod
    def get(cls, source, index=None) -> (__toTuple__, str):
        try:
            if source not in ("backgrounds", "classes", "races", "skills"):
                raise ValueError
        except ValueError:
            exit("source: invalid source specified '{}'.".format(source))
        return cls.__toTuple__(source, index)


class WriterBlobTypeError(TypeError):
    """Raised if data is not of type 'dict'."""


class WriterBlobValueError(Exception):
    """Raised if not all keys have purged_values."""


class Writer:
    def __init__(self, output_file: str, data: dict) -> None:
        if not isinstance(data, dict):
            raise WriterBlobTypeError("data: argument must be of type 'dict'!")
        data_keys = (
            "race",
            "background",
            "class",
            "archetype",
            "level",
            "proficiency",
            "abilities",
            "saves",
            "armors",
            "weapons",
            "tools",
            "slots",
            "languages",
            "skills",
            "feats",
            "equipment",
        )
        if not all(k in data for k in data_keys):
            raise WriterBlobValueError("data: all keys must have a value!")
        blob = list()
        blob.append(data["race"])
        blob.append(data["background"])
        blob.append(data["class"])
        blob.append(data["archetype"])
        blob.append(data["level"])
        blob.append(data["proficiency"])
        blob.append(self._format_ability_scores(data["abilities"]))
        blob.append(
            self._format_saving_throws(data["abilities"], data["level"], data["saves"])
        )
        blob.append(self._format_list(data["armors"]))
        blob.append(self._format_list(data["weapons"]))
        blob.append(self._format_list(data["tools"]))
        blob.append(data["slots"])
        blob.append(self._format_list(data["languages"]))
        blob.append(
            self._format_skills(data["skills"], data["abilities"], data["level"])
        )
        blob.append(self._format_list(data["feats"]))
        blob.append(self._format_list(data["equipment"]))
        self.file = output_file
        self.data = tuple(blob)

    def __enter__(self):
        return self

    def __exit__(self, exec_type, value, traceback):
        pass

    @staticmethod
    def _format_ability_scores(scores: OrderedDict) -> str:
        """Formats ability scores for character sheet."""
        my_scores = ""
        for ability, values in scores.items():
            value = values["Value"]
            modifier = values["Modifier"]
            my_scores += '<{} score="{}" modifier="{}" />'.format(
                ability, value, modifier
            )
        return my_scores

    @staticmethod
    def _format_list(iterable: (list, set, tuple)) -> str:
        """Formats list elements for character sheet."""
        my_iterable = ""
        if len(iterable) < 1:
            return my_iterable
        for entry in iterable:
            entry = entry.replace(" ", "_")
            my_iterable += '<item value="{}" />'.format(entry)
        return my_iterable

    @staticmethod
    def _format_saving_throws(
        scores: OrderedDict, level: int, proficient_saving_throws: (list, set, tuple)
    ) -> str:
        """Formats saving throws for character sheet."""
        saving_throws = (
            "Strength",
            "Dexterity",
            "Constitution",
            "Intelligence",
            "Wisdom",
            "Charisma",
        )
        my_saves = ""
        for saving_throw in saving_throws:
            ability_modifier = scores[saving_throw]["Modifier"]
            proficiency_bonus = (
                saving_throw in proficient_saving_throws
                and get_proficiency_bonus(level)
                or 0
            )
            total_save = ability_modifier + proficiency_bonus
            my_saves += '<{} value="{}"/>'.format(saving_throw, total_save)
        return my_saves

    @staticmethod
    def _format_skills(
        skills: (list, set, tuple), scores: OrderedDict, level: int
    ) -> str:
        """Formats skill values for character sheet."""
        expanded_skills = {}
        for skill in skills:
            skill_ability = Skills.get_skill_ability(skill)
            expanded_skills[skill] = {
                "Ability": skill_ability,
                "Modifier": get_ability_modifier(scores, skill_ability),
            }
        my_skills = ""
        for skill, values in sorted(expanded_skills.items()):
            skill_name = skill.replace(" ", "_")
            skill_total = get_proficiency_bonus(level) + values["Modifier"]
            my_skills += '<{} ability="{}" value="{}" />'.format(
                skill_name, values["Ability"], skill_total,
            )
        return my_skills

    def write(self) -> None:
        """Writes data to character sheet."""
        template = """<?xml version="1.0"?>
            <Personae>
                <Race name="%s" background="%s" />
                <Class name="%s" archetype="%s" level="%s" proficiency="%s" />
                <AbilityScores>%s</AbilityScores>
                <SavingThrows>%s</SavingThrows>
                <ArmorProficiencies>%s</ArmorProficiencies>
                <WeaponProficiencies>%s</WeaponProficiencies>
                <ToolProficiencies>%s</ToolProficiencies>
                <Spell slots="%s" />
                <Languages>%s</Languages>
                <Skills>%s</Skills>
                <Feats>%s</Feats>
                <Equipment>%s</Equipment>
            </Personae>"""
        with open(self.file, "w+") as character_sheet:
            data = template % self.data
            character_sheet.write(BeautifulSoup(data, "xml").prettify())
        character_sheet.close()


def _err(message: str) -> None:
    click.secho(
        message, bold=True, err=True, fg="red",
    )
    exit()


def _out(message: str) -> None:
    click.secho(
        message, bold=True, fg="green",
    )


@click.command()
@click.option(
    "-race",
    default=ValidParameters.get("races", 0),
    show_default=True,
    type=str,
    required=True,
    help="Character's chosen race.",
)
@click.option(
    "-background",
    default=Classes.get_class_background(ValidParameters.get("classes", 0)),
    show_default=True,
    type=str,
    required=True,
    help="Character's chosen background.",
)
@click.option(
    "-class_",
    default=ValidParameters.get("classes", 0),
    show_default=True,
    type=str,
    required=True,
    help="Character's chosen class.",
)
@click.option(
    "-archetype", default=None, type=str, help="Character's chosen archetype."
)
@click.option(
    "-level",
    default=1,
    show_default=True,
    type=int,
    required=True,
    help="Character's chosen class level.",
)
@click.option(
    "-variant",
    default="false",
    show_default=True,
    type=str,
    required=True,
    help="Determines if variant trait applies.",
)
@click.version_option(version=__version__)
def main(
    race: str, background: str, class_: str, archetype: str, level: int, variant=str,
) -> None:
    # Create sources save directory (if necessary).
    home_directory = os.path.expanduser("~/Personae")
    if not os.path.exists(home_directory):
        os.mkdir(home_directory)
        click.echo("Personae directory created at '{}'.".format(home_directory))
    click.echo("Characters will be saved to '{}'.".format(home_directory))

    # CLI argument checking
    if race not in ValidParameters.get("races"):
        _err("Invalid sources race '{}'.".format(race))
    if background not in ValidParameters.get("backgrounds"):
        _err("Invalid sources background '{}'.".format(background))
    if class_ not in ValidParameters.get("classes"):
        _err("Invalid sources class '{}'.".format(class_))
    if archetype is not None and archetype not in Classes.get_archetypes(class_):
        _err("Invalid sources archetype '{}' for '{}'.".format(archetype, class_))
    if level not in range(0, 21):
        _err("Invalid level '{}' specified. Range (1-20).".format(level))
    if variant not in ("false", "true"):
        _err("Invalid variant option '{}' specified. false|true.".format(variant))
    else:
        if variant == "false":
            variant = False
        elif variant == "true":
            variant = True

    # Rudimentary sources additions.
    equipment = Backgrounds.get_background_equipment(background)
    slots = Classes.get_spell_slots(class_, level, archetype)

    # Determine and assign ability score improvements, if applicable.
    a = ScoreAssigner(class_, race, variant)
    asi = ScoreImprovement(
        _class=class_,
        archetype=archetype,
        level=level,
        race=race,
        abilities=a.abilities,
        primaries=a.primaries,
        armor_proficiency=get_proficiency("armors", race, class_),
        tool_proficiency=get_proficiency("tools", race, class_),
        weapon_proficiency=get_proficiency("weapons", race, class_),
        languages=Races.get_languages(race, background),
        skills=Skills.select_skills(race, background, class_, variant),
        feats=list(),
        variant=variant,
    )
    asi.assign_upgrades()
    languages = asi.languages
    armors = asi.armors
    tools = asi.tools
    weapons = asi.weapons
    feats = asi.feats
    skills = asi.skills
    abilities = asi.abilities
    saves = asi.saves

    # Prepare data for writing.
    data = dict()
    data["race"] = race
    data["background"] = background
    data["class"] = class_
    data["archetype"] = archetype
    data["level"] = level
    data["proficiency"] = get_proficiency_bonus(level)
    data["abilities"] = abilities
    data["saves"] = saves
    data["armors"] = armors
    data["tools"] = tools
    data["weapons"] = weapons
    data["slots"] = slots
    data["languages"] = languages
    data["skills"] = skills
    data["feats"] = feats
    data["equipment"] = equipment

    # Character sheet data review.
    _out("*" * 40)
    _out("Race (Background): {} ({})".format(data["race"], data["background"]))
    _out("Class (Level): {} ({})".format(data["class"], level))
    _out("Archetype: {}".format(data["archetype"]))
    for ability, values in data["abilities"].items():
        _out("{}: {} ({})".format(ability, values["Value"], values["Modifier"]))
    _out("Armor Proficiencies: {}".format(", ".join(data["armors"])))
    _out("Tool Proficiencies: {}".format(", ".join(data["tools"])))
    _out("Weapon Proficiencies: {}".format(", ".join(data["weapons"])))
    _out("Languages: {}".format(", ".join(data["languages"])))
    _out("Feats: {}".format(", ".join(data["feats"])))
    _out("Skills: {}".format(", ".join(data["skills"])))
    _out("Saving Throws: {}".format(", ".join(data["saves"])))
    _out("Equipment: {}".format(", ".join(data["equipment"])))
    _out("*" * 40)

    # Prompt.
    while True:
        click.echo("Save character? (y/n) ", nl=False)
        prompt = click.getchar()
        click.echo()
        if prompt in ("y", "Y"):
            try:
                file_name = time.ctime(time.time()).replace(" ", "_") + ".xml"
                fp = "{}/{}".format(home_directory, file_name)
                with Writer(fp, data) as output:
                    output.write()
                click.echo("Character saved to '{}'.".format(fp))
            except WriterBlobValueError as err:
                click.echo(err)
            except WriterBlobTypeError as err:
                click.echo(err)
            break
        elif prompt in ("n", "N"):
            click.echo("Aborted. Character not saved.")
            break
        else:
            continue


if __name__ == "__main__":
    main()



from collections import OrderedDict
from random import choice, sample

import numpy

from personaednd.dice import Dice
from personaednd.listutils import add, merge, pick, purge
from personaednd.sources import (
    get_allotted_feats,
    Classes,
    Feats,
    Races,
    Skills,
    Tools,
    Weapons,
)


class ScoreAssigner:
    def __init__(self, _class: str, race: str, variant: bool) -> None:
        # Initial purged_values.
        raw_scores = list(self.roll_drop_lowest(60))
        primary_abilities = list()
        ability_scores = OrderedDict()
        ability_scores["Strength"] = None
        ability_scores["Dexterity"] = None
        ability_scores["Constitution"] = None
        ability_scores["Intelligence"] = None
        ability_scores["Wisdom"] = None
        ability_scores["Charisma"] = None
        # Assign primary abilities.
        primary_class_abilities = Classes.get_class_primary_abilities(_class)
        for rank, ability in primary_class_abilities.items():
            if isinstance(ability, list):
                ability = choice(ability)
            highest_score = max(raw_scores)
            raw_scores.remove(highest_score)
            modifier = get_ability_modifier(highest_score)
            ability_scores[ability] = {"Value": highest_score, "Modifier": modifier}
            primary_abilities.append(ability)
        # Assign secondary abilities.
        for ability, values in ability_scores.items():
            if values is None:
                chosen_score = choice(raw_scores)
                raw_scores.remove(chosen_score)
                modifier = get_ability_modifier(chosen_score)
                ability_scores[ability] = {"Value": chosen_score, "Modifier": modifier}
        # Apply racial bonus(es).
        bonuses = Races.get_racial_bonus(race, variant)
        for ability, bonus in bonuses.items():
            old_score = ability_scores[ability]["Value"]
            new_score = old_score + bonus
            if new_score > 20:
                new_score = 20
            modifier = get_ability_modifier(new_score)
            ability_scores[ability] = {"Value": new_score, "Modifier": modifier}
        self.abilities = ability_scores
        self.primaries = tuple(primary_abilities)

    @staticmethod
    def roll_drop_lowest(threshold: int) -> numpy.ndarray:
        """Returns six totaled results of 3(4) at threshold or higher."""
        die = Dice("d6", 4)
        results = numpy.array([], dtype=int)
        while results.sum() < threshold:
            for _ in range(1, 7):
                result = die.roll()
                excl_result = result.min(initial=None)
                ability_score = result.sum() - excl_result
                results = numpy.append(results, ability_score)
            score_total = results.sum()
            try:
                maximum_score = results.max(initial=None)
                minimum_score = results.min(initial=None)
            except ValueError:
                # Empty array FIX, force a re-roll.
                maximum_score = 14
                minimum_score = 7
            if score_total < threshold or maximum_score < 15 or minimum_score < 8:
                results = numpy.array([], dtype=int)
        return results


class ScoreImprovement:
    def __init__(
        self,
        _class: str,
        archetype: str,
        level: int,
        race: str,
        abilities: OrderedDict,
        primaries: tuple,
        armor_proficiency: list,
        tool_proficiency: list,
        weapon_proficiency: list,
        languages: list,
        skills: list,
        feats: list,
        variant: bool,
    ) -> None:
        # Character data
        self._class = _class
        self.archetype = archetype
        self.level = level
        self.race = race
        self.abilities = abilities
        self.primaries = primaries
        self.armors = armor_proficiency
        self.tools = tool_proficiency
        self.weapons = weapon_proficiency
        self.languages = languages
        self.skills = skills
        self.feats = feats
        self.variant = variant
        self.saves = Classes.get_saving_throws(self._class)
        # Humans w/ variant rule get a bonus feat.
        if race == "Human" and variant:
            self._upgrade_feats()

    def _assign_features(self, feat) -> None:
        """Assign associated features by specified feat."""
        # Actor
        if feat == "Actor":
            self._set_ability("Charisma", 1)
        # Athlete/Lightly Armored/Moderately Armored/Weapon Master
        if feat in (
            "Athlete",
            "Lightly Armored",
            "Moderately Armored",
            "Weapon Master",
        ):
            ability_choice = choice(["Strength", "Dexterity"])
            self._set_ability(ability_choice, 1)
            if feat == "Lightly Armored":
                add(self.armors, "Light")
            elif feat == "Moderately Armored":
                add(self.armors, "Medium")
                add(self.armors, "Shield")
        # Durable
        if feat == "Durable":
            self._set_ability("Constitution", 1)
        # Heavily Armored/Heavy Armor Master
        if feat in ("Heavily Armored", "Heavy Armor Master"):
            self._set_ability("Strength", 1)
            if feat == "Heavily Armored":
                add(self.armors, "Heavy")
        # Keen Mind/Linguist
        if feat in ("Keen Mind", "Linguist"):
            self._set_ability("Intelligence", 1)
            if feat == "Linguist":
                # Remove already known languages.
                linguist_languages = [
                    "Abyssal",
                    "Celestial",
                    "Common",
                    "Deep Speech",
                    "Draconic",
                    "Dwarvish",
                    "Elvish",
                    "Giant",
                    "Gnomish",
                    "Goblin",
                    "Halfling",
                    "Infernal",
                    "Orc",
                    "Primordial",
                    "Sylvan",
                    "Undercommon",
                ]
                purge(linguist_languages, self.languages)
                # Choose 3 bonus languages.
                merge(self.languages, sample(linguist_languages, 3))
        # Observant
        if feat == "Observant":
            if self._class in ("Cleric", "Druid"):
                self._set_ability("Wisdom", 1)
            elif self._class in ("Wizard",):
                self._set_ability("Intelligence", 1)
        # Resilient
        if feat == "Resilient":
            # Remove all proficient saving throws.
            resilient_saves = [
                "Strength",
                "Dexterity",
                "Constitution",
                "Intelligence",
                "Wisdom",
                "Charisma",
            ]
            purge(resilient_saves, self.saves)
            # Choose one non-proficient saving throw.
            ability_choice = choice(resilient_saves)
            self._set_ability(ability_choice, 1)
            add(self.saves, ability_choice)
        # Skilled
        if feat == "Skilled":
            for _ in range(3):
                skilled_choice = choice(["Skill", "Tool"])
                if skilled_choice == "Skill":
                    skills = purge(Skills.get_skills(), self.skills)
                    add(self.skills, choice(skills))
                elif skilled_choice == "Tool":
                    tools = purge(Tools.get_tools(None), self.tools)
                    add(self.tools, choice(tools))
        # Tavern Brawler
        if feat == "Tavern Brawler":
            self._set_ability(choice(["Strength", "Constitution"]), 1)
            add(self.weapons, "Improvised weapons")
            add(self.weapons, "Unarmed strikes")
        # Weapon Master
        if feat == "Weapon Master":
            weapons = Weapons.get_martial_weapons()
            if "Simple" not in self.weapons:
                merge(weapons, Weapons.get_simple_weapons())
            for weapon in self.weapons:
                if weapon in ("Simple", "Martial"):
                    continue
                if weapon in weapons:
                    weapons.remove(weapon)
            bonus_weapons = sample(weapons, 4)
            for weapon in bonus_weapons:
                add(self.weapons, weapon)

    def _isadjustable(self, abilities: (list, set, str, tuple), bonus: int) -> bool:
        """Checks if abilities is lt or eq to 20 with bonus."""
        if isinstance(abilities, (list, set, tuple)):
            for ability in abilities:
                score = self.abilities[ability]["Value"] + bonus
                if score > 20:
                    return False
            return True
        elif isinstance(abilities, str):
            score = self.abilities[abilities]["Value"] + bonus
            return score <= 20 and True or False
        return True

    def _set_ability(self, ability_name: str, bonus: int) -> None:
        if self._isadjustable(ability_name, bonus):
            old = self.abilities[ability_name]["Value"]
            new = old + bonus
            self.abilities[ability_name]["Value"] = new
            self.abilities[ability_name]["Modifier"] = get_ability_modifier(new)

    def _upgrade_ability(self) -> None:
        """Assigns ability upgrade or adds a feat (if not applicable)."""
        bonus = choice([1, 2])
        try:
            if bonus is 2:
                if self._isadjustable(self.primaries, bonus):
                    for ability in self.primaries:
                        if self._isadjustable(ability, bonus):
                            self._set_ability(ability, bonus)
                            break
                else:
                    raise ValueError
            elif bonus is 1:
                if all(
                    self._isadjustable(ability, bonus) is True
                    for ability in self.primaries
                ):
                    for ability in self.primaries:
                        self._set_ability(ability, bonus)
                else:
                    raise ValueError
        except ValueError:
            self._upgrade_feats()

    def _upgrade_feats(self) -> None:
        """Assigns a random valid feat."""
        feats = purge(Feats.get_feats(), self.feats)
        feat = pick(feats)
        has_required = False
        while True:
            if has_required:
                break
            while True:
                if Feats.has_requirements(
                    feat=feat,
                    _class=self._class,
                    archetype=self.archetype,
                    level=self.level,
                    abilities=self.abilities,
                    armor_proficiency=self.armors,
                    weapon_proficiency=self.weapons,
                ):
                    has_required = True
                    self.feats.append(feat)
                else:
                    feat = pick(feats)
                break
        self.feats.sort()
        self._assign_features(feat)

    def assign_upgrades(self):
        num_of_feats = get_allotted_feats(self._class, self.level)
        if num_of_feats > 0:
            for upgrade in range(num_of_feats):
                upgrade_selection = choice(("Ability", "Feat"))
                if upgrade_selection == "Ability":
                    self._upgrade_ability()
                elif upgrade_selection == "Feat":
                    self._upgrade_feats()


def get_ability_modifier(score) -> int:
    """Returns ability modifier by score."""
    return score is not 0 and int((score - 10) / 2) or 0



from collections import OrderedDict
from math import ceil
from random import choice, sample

from personaednd.listutils import merge, pick, purge
from personaednd.sources.reader import Reader


class Backgrounds(Reader):
    def __init__(self) -> None:
        super().__init__("backgrounds")

    @classmethod
    def get_backgrounds(cls) -> tuple:
        """Returns a tuple of available backgrounds."""
        return tuple(cls().find().keys())

    @classmethod
    def get_background_equipment(cls, background: str) -> list:
        """Returns a list of bonus equipment by background."""
        return cls().find([background, "equipment"])

    @classmethod
    def get_background_languages(cls, background: str) -> int:
        """Returns a count of bonus languages by background."""
        return cls().find([background, "languages"])

    @classmethod
    def get_background_skills(cls, background: str) -> list:
        """Returns a list of bonus skills by background."""
        return cls().find([background, "skills"])

    @classmethod
    def get_background_tools(cls, background: str) -> list:
        """Returns a list of bonus tool proficiencies by background."""
        background_tools = list()
        for tool in cls().find([background, "tool_proficiency"]):
            if Tools.has_sub_tools(tool):
                sub_tool = choice(Tools.get_sub_tools(tool))
                background_tools.append("{}: {}".format(tool, sub_tool))
            else:
                background_tools.append(tool)
        return background_tools


class Classes(Reader):
    def __init__(self) -> None:
        super().__init__("classes")

    @classmethod
    def get_archetypes(cls, _class: str) -> list:
        """Returns a list of archetypes by class."""
        return cls().find([_class, "archetypes"])

    @classmethod
    def get_class_armor_proficiency(cls, _class: str) -> list:
        """Returns a list of armor proficiencies by class."""
        return cls().find([_class, "armor_proficiency"])

    @classmethod
    def get_class_background(cls, _class: str) -> str:
        """Returns default class background by class."""
        return cls().find([_class, "background"])

    @classmethod
    def get_class_primary_abilities(cls, _class: str) -> dict:
        """Returns a dict of primary abilities by class."""
        return cls().find([_class, "primary_abilities"])

    @classmethod
    def get_class_tool_proficiency(cls, _class: str) -> list:
        """Returns a list of tool proficiencies by class."""
        if _class == "Monk":
            return [choice(Tools.get_tools(["Artisan's tools", "Musical instrument"]))]
        else:
            return cls().find([_class, "tool_proficiency"])

    @classmethod
    def get_class_weapon_proficiency(cls, _class: str) -> list:
        """Returns a list of weapon proficiencies by class."""
        return cls().find([_class, "weapon_proficiency"])

    @classmethod
    def get_classes(cls) -> tuple:
        """Returns a tuple of available classes."""
        return tuple(cls().find().keys())

    @classmethod
    def get_hit_die(cls, _class: str) -> int:
        """Returns a hit die integer value by class."""
        return cls().find([_class, "hit_die"])

    @classmethod
    def get_saving_throws(cls, _class: str) -> list:
        """Returns a list of saving throws by class."""
        return cls().find([_class, "saving_throws"])

    @classmethod
    def get_spell_slots(cls, _class: str, level: int, archetype: str) -> str:
        """Returns spell slots by class, level and archetype."""
        try:
            if _class == "Fighter" and archetype != "Eldritch Knight":
                raise KeyError
            elif _class == "Rogue" and archetype != "Arcane Trickster":
                raise KeyError
            else:
                spell_slots = cls().find([_class, "spell_slots", level])
                if len(spell_slots) is not 0:
                    return spell_slots
                else:
                    raise KeyError
        except KeyError:
            return ""


class Feats(Reader):
    def __init__(self) -> None:
        super().__init__("feats")

    @classmethod
    def get_ability_requirements(cls, feat: str) -> dict:
        """Returns ALL ability score requirements for feat."""
        requirements = {}
        strength = cls().find([feat, "strength"])
        if strength > 1:
            requirements["Strength"] = strength
        dexterity = cls().find([feat, "dexterity"])
        if dexterity > 1:
            requirements["Dexterity"] = dexterity
        constitution = cls().find([feat, "constitution"])
        if constitution > 1:
            requirements["Constitution"] = constitution
        intelligence = cls().find([feat, "intelligence"])
        if intelligence > 1:
            requirements["Intelligence"] = intelligence
        wisdom = cls().find([feat, "wisdom"])
        if wisdom > 1:
            requirements["Wisdom"] = wisdom
        charisma = cls().find([feat, "charisma"])
        if charisma > 1:
            requirements["Charisma"] = charisma
        return requirements

    @classmethod
    def get_caster_requirement(cls, feat: str) -> dict:
        """Returns caster requirements for feat."""
        return cls().find([feat, "caster"])

    @classmethod
    def get_feats(cls) -> list:
        """Returns a list of feats."""
        return list(cls().find().keys())

    @classmethod
    def get_proficiency_requirement(cls, feat: str) -> tuple:
        """Returns a tuple of proficiency requirements for feat."""
        return tuple(cls().find([feat, "proficiency"]))

    @classmethod
    def has_requirements(
        cls,
        feat: str,
        _class: str,
        archetype: str,
        level: int,
        abilities: OrderedDict,
        armor_proficiency: list,
        weapon_proficiency: list,
    ) -> bool:
        """Checks if all requirements for feat are met."""
        if (
            feat in ("Heavily Armored", "Lightly Armored", "Moderately Armored",)
            and _class == "Monk"
        ):
            return False
        elif feat in (
            "Heavily Armored",
            "Lightly Armored",
            "Moderately Armored",
            "Weapon Master",
        ):
            # Character already has heavy armor proficiency.
            if feat == "Heavily Armored" and "Heavy" in armor_proficiency:
                return False
            # Character already has light armor proficiency.
            if feat == "Lightly Armored" and "Light" in armor_proficiency:
                return False
            # Character already has medium armor proficiency.
            if feat == "Moderately Armored" and "Medium" in armor_proficiency:
                return False
            # Character already has martial weapon proficiency.
            if feat == "Weapon Master" and "Martial" in weapon_proficiency:
                return False
        elif feat == "Magic Initiative" and _class not in (
            "Bard",
            "Cleric",
            "Druid",
            "Sorcerer",
            "Warlock",
            "Wizard",
        ):
            return False
        proficiency = cls().get_proficiency_requirement(feat)
        proficiencies = armor_proficiency + weapon_proficiency
        if len(proficiency) is not 0 and proficiency not in proficiencies:
            return False
        if cls().get_caster_requirement(feat) and not is_caster(
            _class, level, archetype
        ):
            return False
        for name, value in cls().get_ability_requirements(feat).items():
            my_value = abilities[name]["Value"]
            if feat == "Ritual Caster":
                if my_value >= value:
                    return True
            else:
                if my_value < value:
                    return False
        return True


class Races(Reader):
    def __init__(self) -> None:
        super().__init__("races")

    @classmethod
    def get_languages(cls, race: str, background: str) -> list:
        """Returns a list of languages by race and background."""
        languages = cls().find([race, "languages"])
        # Remove duplicate languages.
        other_languages = [
            "Dwarvish",
            "Elvish",
            "Giant",
            "Gnomish",
            "Goblin",
            "Halfling",
            "Orc",
        ]
        purge(other_languages, languages)
        # Add racial bonus language, if applicable.
        if race in ("Elf, High", "Half-elf", "Human"):
            languages.append(pick(other_languages))
        # Add background bonus language(s).
        background_languages = Backgrounds.get_background_languages(background)
        if background_languages > 0:
            merge(languages, sample(other_languages, background_languages))
        languages.sort()
        return languages

    @classmethod
    def get_races(cls) -> tuple:
        """Returns a tuple of available races."""
        return tuple(cls().find().keys())

    @classmethod
    def get_racial_armor_proficiency(cls, race: str) -> list:
        """Returns a list of armor proficiencies by race."""
        return cls().find([race, "armor_proficiency"])

    @classmethod
    def get_racial_bonus(cls, race: str, variant: bool) -> dict:
        """Returns a dictionary of ability bonuses by race."""
        ability = cls().find([race])
        bonuses = {}
        if ability["strength"] is not 0:
            bonuses["Strength"] = ability["strength"]
        if ability["dexterity"] is not 0:
            bonuses["Dexterity"] = ability["dexterity"]
        if ability["constitution"] is not 0:
            bonuses["Constitution"] = ability["constitution"]
        if ability["intelligence"] is not 0:
            bonuses["Intelligence"] = ability["intelligence"]
        if ability["wisdom"] is not 0:
            bonuses["Wisdom"] = ability["wisdom"]
        if ability["charisma"] is not 0:
            bonuses["Charisma"] = ability["charisma"]
        # Half-elf or Human w/ variant rule.
        if race == "Half-elf" or race == "Human" and variant:
            variant_abilities = [
                "Strength",
                "Dexterity",
                "Constitution",
                "Intelligence",
                "Wisdom",
                "Charisma",
            ]
            # Half-elves get a +2 Charisma, omit this from choices.
            if race == "Half-elf":
                variant_abilities.remove("Charisma")
            abilities = sample(variant_abilities, 2)
            for ability in abilities:
                bonuses[ability] = 1
        # Human w/o variant rule.
        elif race == "Human":
            bonuses["Strength"] = 1
            bonuses["Dexterity"] = 1
            bonuses["Constitution"] = 1
            bonuses["Intelligence"] = 1
            bonuses["Wisdom"] = 1
            bonuses["Charisma"] = 1
        return bonuses

    @classmethod
    def get_racial_tool_proficiency(cls, race: str) -> list:
        """Returns a list of tool proficiencies by race."""
        racial_tools = cls().find([race, "tool_proficiency"])
        if race in ("Dwarf", "Dwarf, Hill", "Dwarf, Mountain"):
            return [choice(racial_tools)]
        return racial_tools

    @classmethod
    def get_racial_weapon_proficiency(cls, race: str) -> list:
        """Returns a list of weapon proficiencies by race."""
        return cls().find([race, "weapon_proficiency"])


class Skills(Reader):
    def __init__(self) -> None:
        super().__init__("skills")

    @classmethod
    def get_class_skills(cls, _class: str) -> list:
        """Returns ALL skills by class."""
        skill_list = list()
        for name in list(cls().find().keys()):
            if _class in cls().get_proficient_classes(name):
                skill_list.append(name)
        return skill_list

    @classmethod
    def get_proficient_classes(cls, skill: str) -> tuple:
        """Returns a tuple of ALL proficient classes by skill."""
        return tuple(cls().find([skill, "classes"]))

    @classmethod
    def get_skill_ability(cls, skill: str) -> str:
        """Returns a skill's associated ability name by skill."""
        return cls().find([skill, "ability"])

    @classmethod
    def get_skills(cls) -> list:
        """Returns ALL skills."""
        return list(cls().find().keys())

    @classmethod
    def select_skills(
        cls, race: str, background: str, _class: str, variant: bool
    ) -> list:
        """Randomly selects skills based upon race, background and class."""
        skills = list()
        background_skills = Backgrounds.get_background_skills(background)
        class_skills = cls().get_class_skills(_class)
        # Add racial skills first.
        if race in ("Elf", "Elf, Drow", "Elf, High", "Elf, Wood", "Half-orc"):
            if race == "Half-orc":
                racial_skill = "Intimidation"
            else:
                racial_skill = "Perception"
            skills.append(racial_skill)
        elif race == "Half-elf" or race == "Human":
            # Remove background, class skills from all skill choices.
            all_skill_choices = cls().get_skills()
            purge(all_skill_choices, background_skills)
            purge(all_skill_choices, class_skills)
            if race == "Half-elf":
                merge(skills, sample(all_skill_choices, 2))
            elif race == "Human" and variant:
                skills.append(choice(all_skill_choices))
        # Next, add background skills.
        if len(background_skills) is not 0:
            merge(skills, background_skills)
        # Finally, choose class skills.
        purge(class_skills, skills)
        merge(skills, sample(class_skills, get_skill_allotment(_class)))
        skills.sort()
        return skills


class Tools(Reader):
    def __init__(self) -> None:
        super().__init__("tools")

    @classmethod
    def get_main_tools(cls) -> tuple:
        """Returns a tuple of all main level tools."""
        return tuple(cls().find().keys())

    @classmethod
    def get_sub_tools(cls, tool: str) -> tuple:
        """Return a tuple of sub level tools for a tool."""
        tools = tuple(cls().find([tool]))
        if len(tools) > 0:
            return tools
        return tuple()

    @classmethod
    def get_tools(cls, main_tools: (list, set, tuple, None)) -> list:
        """Returns a tuple of ALL tools or main tools, if applicable."""
        tools = list()
        if isinstance(main_tools, (list, set, tuple)):
            for tool in main_tools:
                if not cls().has_sub_tools(tool):
                    continue
                for sub_tool in cls().get_sub_tools(tool):
                    tools.append("{}: {}".format(tool, sub_tool))
        else:
            for tool in list(cls().get_main_tools()):
                if cls().has_sub_tools(tool):
                    for sub_tool in cls().get_sub_tools(tool):
                        tools.append("{}: {}".format(tool, sub_tool))
                else:
                    tools.append(tool)
        tools.sort()
        return tools

    @classmethod
    def has_sub_tools(cls, tool: str) -> bool:
        """Checks if a tool has sub level tools."""
        if len(cls().get_sub_tools(tool)) is 0:
            return False
        return True


class Weapons(Reader):
    def __init__(self) -> None:
        super().__init__("weapons")

    @classmethod
    def get_martial_weapons(cls) -> list:
        """Returns a list of martial weapons."""
        return cls().find(["Martial"])

    @classmethod
    def get_simple_weapons(cls) -> list:
        """Returns a list of simple weapons."""
        return cls().find(["Simple"])


def get_ability_modifier(scores, ability) -> int:
    """Returns ability modifier value from dictionary of abilities."""
    return scores[ability]["Modifier"]


def get_skill_allotment(_class: str) -> int:
    """Returns allotment value of skills by class."""
    if _class in ("Rogue",):
        skill_allotment = 4
    elif _class in ("Bard", "Ranger"):
        skill_allotment = 3
    else:
        skill_allotment = 2
    return skill_allotment


def get_allotted_feats(_class: str, level: int) -> int:
    """Returns allotment value of feats by class and level."""
    allotted_feats = 0
    for _ in range(1, level + 1):
        if (_ % 4) == 0 and _ is not 20:
            allotted_feats += 1
    if _class == "Fighter" and level >= 6:
        allotted_feats += 1
    if _class == "Rogue" and level >= 8:
        allotted_feats += 1
    if _class == "Fighter" and level >= 14:
        allotted_feats += 1
    if level >= 19:
        allotted_feats += 1
    return allotted_feats


def get_proficiency(proficiency: str, _race: str, _class: str) -> list:
    proficiency_list = {
        "armors": {
            "class": Classes.get_class_armor_proficiency(_class),
            "racial": Races.get_racial_armor_proficiency(_race),
        },
        "tools": {
            "class": Classes.get_class_tool_proficiency(_class),
            "racial": Races.get_racial_tool_proficiency(_race),
        },
        "weapons": {
            "class": Classes.get_class_weapon_proficiency(_class),
            "racial": Races.get_racial_weapon_proficiency(_race),
        },
    }
    try:
        proficiencies = proficiency_list[proficiency]["racial"]
        proficiencies += proficiency_list[proficiency]["class"]
        proficiencies = list(dict.fromkeys(proficiencies))
        proficiencies.sort()
        return proficiencies
    except KeyError:
        return list()


def get_proficiency_bonus(level: int) -> int:
    """Returns a proficiency bonus value by level."""
    return ceil((level / 4) + 1)


def is_caster(_class: str, level: int, archetype=None) -> bool:
    """Returns True if caster by _class, level, archetype or False if not."""
    if _class in ("Bard", "Cleric", "Druid", "Sorcerer", "Warlock", "Wizard"):
        return True
    elif _class in ("Paladin", "Ranger") and level > 1:
        return True
    elif _class is "Fighter" and archetype in ("Eldritch Knight",):
        return True
    elif _class is "Rogue" and archetype in ("Arcane Trickster",):
        return True
    return False



from random import choice


def add(iterable: list, value: (float, int, str)) -> list:
    """Adds a value into iterable."""
    iterable.append(value)
    iterable.sort()
    return iterable


def merge(original_list: list, merged_list_values: (list, tuple)) -> list:
    """Merges individual unique merged_list_values into original_list."""
    for value in merged_list_values:
        if value not in original_list:
            original_list.append(value)
    return original_list


def pick(iterable: list) -> (bool, float, int, str):
    """Chooses random value from list then removes it."""
    try:
        if not isinstance(iterable, list):
            raise TypeError
        if len(iterable) is 0:
            raise ValueError
        selection = choice(iterable)
        iterable.remove(selection)
        return selection
    except (TypeError, ValueError):
        return False


def purge(original_list: list, purged_list_values: (list, tuple)) -> list:
    """Purges individual unique purged_list_values from original_list."""
    for value in purged_list_values:
        if value in original_list:
            original_list.remove(value)
    return original_list



from numpy.random import randint
import numpy


class Dice:
    def __init__(self, sides: str, rolls: int) -> None:
        dice_bag = {
            "d4": 4,
            "d6": 6,
            "d8": 8,
            "d10": 10,
            "d12": 12,
            "d20": 20,
            "d100": 100,
        }
        try:
            self.sides = dice_bag[sides]
        except KeyError:
            self.sides = dice_bag["d4"]
        self.sides += 1
        self.rolls = rolls >= 1 and rolls or 1

    def roll(self) -> numpy.ndarray:
        """Returns random array purged_values between low and high at size."""
        return randint(low=1, high=self.sides, size=self.rolls)



from setuptools import find_packages, setup

from personaednd.version import __version__


setup(
    name="personaednd",
    version=__version__,
    packages=find_packages(),
    package_data={"personaednd": ["sources/*.yml"]},
    url="https://github.com/sg679/personaednd",
    license="MIT",
    author="Marcus T Taylor",
    description="Randomly generates 5th edition Dungeons & Dragons characters.",
    long_description=open("README.md", "r").read(),
    long_description_content_type="text/markdown",
    install_requires=[
        "beautifulsoup4>=4.8",
        "click>=7.0",
        "lxml>=4.4",
        "numpy>=1.17",
        "PyYAML>=5.1",
    ],
    python_requires=">=3.0",
    entry_points={"console_scripts": ["personaednd=personaednd.shell:main"]},
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Games/Entertainment :: Role-Playing",
    ],
)
